from typing import Any, Dict, Iterable, Type, Union

CONTEXT_HEADER_KEY = "x-spymaster-context"
CONTEXT_ID_HEADER_KEY = "x-spymaster-context-id"
JsonType = Union[str, int, float, bool, list, Dict[str, Any], None]
ErrorTypes = Iterable[Type[Exception]]



from logging.config import dictConfig

from the_spymaster_util.logger import get_dict_config


def configure_logging():
    config = get_dict_config(std_formatter="debug")
    dictConfig(config)


configure_logging()



import json
import logging
from typing import Any, Callable, Dict, Mapping, Optional, Type, Union

import requests
from requests import HTTPError, Response
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from the_spymaster_util.http.defs import CONTEXT_HEADER_KEY, ErrorTypes
from the_spymaster_util.http.errors import DEFAULT_ERRORS
from the_spymaster_util.logger import wrap
from the_spymaster_util.measure_time import MeasureTime
from the_spymaster_util.strings import camel_to_const, get_class_name

log = logging.getLogger(__name__)

DEFAULT_RETRY_STRATEGY = Retry(
    raise_on_status=False,
    total=2,
    backoff_factor=0.3,
    status_forcelist=[429, 500, 502, 503, 504],
    allowed_methods=["HEAD", "OPTIONS", "GET", "POST", "PUT", "DELETE"],
)


class BaseHttpClient:
    def __init__(self, base_url: str, retry_strategy: Optional[Retry] = DEFAULT_RETRY_STRATEGY):
        self.base_url = base_url
        self.session = requests.Session()
        self.set_retry_strategy(retry_strategy)

    def set_retry_strategy(self, retry_strategy: Optional[Retry]):
        retry_adapter = HTTPAdapter(max_retries=retry_strategy)
        self.session.mount("http://", retry_adapter)
        self.session.mount("https://", retry_adapter)

    def _http_call(
        self,
        endpoint: str,
        method: Callable,
        parse: bool = True,
        error_types: Optional[ErrorTypes] = DEFAULT_ERRORS,
        **kwargs,
    ) -> Union[dict, Response]:
        url = f"{self.base_url}/{endpoint}"
        headers = kwargs.pop("headers", None) or {}
        data = kwargs.get("data")
        log_context = getattr(log, "context", None)
        if log_context:
            headers[CONTEXT_HEADER_KEY] = json.dumps(log_context)
        log_http_data = kwargs.pop("log_http_data", True)
        method_name = method.__name__.upper()
        _log_request(method=method_name, url=url, data=data, headers=headers, log_http_data=log_http_data)
        with MeasureTime() as mt:
            response = method(url, headers=headers, **kwargs)
        _log_response(method=method_name, url=url, response=response, duration=mt.delta, log_http_data=log_http_data)
        _validate(response=response, error_types=error_types)
        if parse:
            return response.json()
        return response

    def _get(
        self,
        endpoint: str,
        data: dict,
        error_types: Optional[ErrorTypes] = DEFAULT_ERRORS,
        **kwargs,
    ) -> Union[dict, Response]:
        return self._http_call(
            endpoint=endpoint,
            method=self.session.get,
            params=data,
            error_types=error_types,
            **kwargs,
        )

    def _post(
        self,
        endpoint: str,
        data: dict,
        error_types: Optional[ErrorTypes] = DEFAULT_ERRORS,
        **kwargs,
    ) -> Union[dict, Response]:
        return self._http_call(
            endpoint=endpoint,
            method=self.session.post,
            json=data,
            error_types=error_types,
            **kwargs,
        )  # type: ignore


def _validate(response: Response, error_types: Optional[ErrorTypes] = None):
    try:
        response.raise_for_status()
    except HTTPError as http_error:
        _raise_custom_error(response=response, http_error=http_error, error_types=error_types)


def _raise_custom_error(response: Response, http_error: HTTPError, error_types: Optional[ErrorTypes] = None):
    if not error_types:
        raise http_error
    try:
        data = response.json()
    except Exception:  # noqa
        raise http_error
    error_code = data.pop("error_code", None)
    if not error_code:
        raise http_error
    error_code_mapping = _get_error_code_mapping(error_types=error_types)
    error_class = error_code_mapping.get(error_code)
    if not error_class:
        raise http_error
    try:
        parsed_error = error_class(**data)  # type: ignore
    except Exception as error_init_error:
        log.warning(f"Error code {error_code} exists, but failed to init: {error_init_error}")
        raise http_error
    raise parsed_error from http_error


def _log_request(method: str, url: str, data: Optional[dict], headers: Optional[dict], log_http_data: bool = True):
    extra: Dict[str, Any] = {"method": method, "url": url}
    if log_http_data:
        extra["data"] = data
        extra["headers"] = headers
    log.debug(f"Sending: {wrap(method)} to {wrap(url)}", extra=extra)


def _log_response(method: str, url: str, response: Response, duration: float, log_http_data: bool = True):
    extra = {"method": method, "url": url, "status_code": response.status_code, "duration": duration}
    if log_http_data:
        try:
            data = response.json()
        except Exception:  # noqa
            data = str(response.content)
        extra["data"] = data
    log.debug(f"Received: {wrap(response.status_code)}", extra=extra)


def extract_context(headers: Mapping[str, Any]) -> dict:
    try:
        context_json = headers.get(CONTEXT_HEADER_KEY)
        if not context_json:
            return {}
        return json.loads(context_json)
    except Exception as e:
        log.debug(f"Failed to extract context from headers: {e}")
        return {}


def get_error_code(exception: Union[Exception, Type[Exception]]) -> str:
    error_code = getattr(exception, "error_code", None)
    if not error_code:
        error_code = get_class_name(exception)
    return camel_to_const(error_code)


def _get_error_code_mapping(error_types: ErrorTypes) -> Dict[str, Type[Exception]]:
    return {get_error_code(error): error for error in error_types}



import time
from collections import OrderedDict
from dataclasses import dataclass
from threading import RLock
from typing import Any, Optional

# Based on https://github.com/mobilityhouse/ttldict


@dataclass
class TTLItem:
    value: Any
    expire_ts: float


class TTLDict(OrderedDict):
    def __init__(self, default_ttl: float, *args, **kwargs):
        """
        :param default_ttl: Default TTL for items, in seconds.
        """
        self._default_ttl = default_ttl
        self._lock = RLock()
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return "<default_ttl=%r, items=%r;>" % (self._default_ttl, self.items())

    def __len__(self):
        with self._lock:
            self._purge()
            return super().__len__()

    def __contains__(self, item: Any) -> bool:
        with self._lock:
            if not super().__contains__(item):
                return False
            return not self.is_expired(item)

    def __iter__(self):
        for key in super().__iter__():
            if not self.is_expired(key):
                yield key

    def __getitem__(self, key: Any) -> Any:
        with self._lock:
            if self.is_expired(key):
                self.__delitem__(key)
                raise KeyError
            item: TTLItem = super().__getitem__(key)
            return item.value

    def __setitem__(self, key: Any, value: Any):
        self.set(key=key, value=value, ttl=self._default_ttl)

    def __delitem__(self, key: Any):
        with self._lock:
            super().__delitem__(key)

    def get(self, key: Any, default: Any = None) -> Any:
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def set(self, key: Any, value: Any, ttl: Optional[float] = None):
        if not ttl:
            ttl = self._default_ttl
        with self._lock:
            expire_ts = time.time() + ttl
            ttl_item = TTLItem(value=value, expire_ts=expire_ts)
            super().__setitem__(key, ttl_item)

    def is_expired(self, key: Any, now: Optional[float] = None) -> bool:
        with self._lock:
            if now is None:
                now = time.time()
            ttl_item: TTLItem = super().__getitem__(key)
            return ttl_item.expire_ts < now

    def _purge(self):
        now = time.time()
        all_keys = list(super().keys())
        for key in all_keys:
            if self.is_expired(key, now=now):
                self.__delitem__(key)

    def keys(self):
        for key in super().keys():
            if not self.is_expired(key):
                yield key

    def values(self):
        for key, ttl_item in super().items():
            if not self.is_expired(key):
                yield ttl_item.value

    def items(self):
        for key, ttl_item in super().items():
            if not self.is_expired(key):
                yield key, ttl_item.value

    def double_lock_test(self):
        with self._lock:
            print("1")
            with self._lock:
                print("double lock test")

    # def set_ttl(self, key, ttl: int, now=None):
    #     """Set TTL for the given key"""
    #     if now is None:
    #         now = time.time()
    #     with self._lock:
    #         value = self[key]
    #         super().__setitem__(key, (now + ttl, value))
    #
    # def get_ttl(self, key, now=None):
    #     """Return remaining TTL for a key"""
    #     if now is None:
    #         now = time.time()
    #     with self._lock:
    #         expire, _value = super().__getitem__(key)
    #         return expire - now
    #
    # def expire_at(self, key, timestamp):
    #     """Set the key expire timestamp"""
    #     with self._lock:
    #         value = self.__getitem__(key)
    #         super().__setitem__(key, (timestamp, value))



import time
from threading import Thread

from the_spymaster_util.ttl_dict import TTLDict

TTL_DEFAULT = 0.01
TTL_LONG = TTL_DEFAULT * 5

SLEEP_DEFAULT = TTL_DEFAULT * 2
SLEEP_LONG = TTL_LONG * 2


def test_get():
    ttl_dict = TTLDict(TTL_DEFAULT)
    ttl_dict["a"] = 1
    assert ttl_dict.get("a") == 1
    assert ttl_dict.get("b") is None
    assert ttl_dict.get("b", "default") == "default"
    assert len(ttl_dict) == 1
    time.sleep(SLEEP_DEFAULT)
    assert ttl_dict.get("a") is None
    assert ttl_dict.get("a", "default") == "default"
    assert len(ttl_dict) == 0


def test_set():
    ttl_dict = TTLDict(TTL_DEFAULT)
    ttl_dict["a"] = 1
    ttl_dict.set("b", 2, ttl=TTL_LONG)
    assert ttl_dict["a"] == 1
    assert ttl_dict["b"] == 2
    assert len(ttl_dict) == 2
    time.sleep(SLEEP_DEFAULT)
    assert ttl_dict.get("a") is None
    assert ttl_dict.get("b") == 2
    assert len(ttl_dict) == 1
    time.sleep(SLEEP_LONG - SLEEP_DEFAULT)
    assert ttl_dict.get("b") is None
    assert len(ttl_dict) == 0


def test_contains():
    ttl_dict = TTLDict(TTL_DEFAULT)
    ttl_dict["a"] = 1
    assert "a" in ttl_dict
    assert "b" not in ttl_dict
    time.sleep(SLEEP_DEFAULT)
    assert "a" not in ttl_dict
    assert "b" not in ttl_dict


def test_is_expired():
    ttl_dict = TTLDict(TTL_DEFAULT, a=1, b=2)
    assert not ttl_dict.is_expired("a")
    assert not ttl_dict.is_expired("b")
    assert len(ttl_dict) == 2
    time.sleep(SLEEP_DEFAULT)
    assert ttl_dict.is_expired("a")
    assert ttl_dict.is_expired("b")
    assert len(ttl_dict) == 0


def test_len():
    ttl_dict = TTLDict(SLEEP_DEFAULT)
    assert len(ttl_dict) == 0
    ttl_dict["a"] = 1
    ttl_dict["b"] = 2
    assert len(ttl_dict) == 2
    time.sleep(SLEEP_DEFAULT)
    assert len(ttl_dict) == 0


def test_iter_empty():
    ttl_dict = TTLDict(TTL_DEFAULT)
    for key in ttl_dict:
        assert False, f"Iterating empty dictionary gave a key {key}"


def test_iter():
    ttl_dict = TTLDict(TTL_DEFAULT)
    orig_dict = {"a": 1, "b": 2, "c": 3}
    ttl_dict.update(orig_dict)
    assert len(ttl_dict) == len(orig_dict)
    for key in ttl_dict:
        assert ttl_dict[key] == orig_dict[key]


def test_keys():
    ttl_dict = TTLDict(TTL_DEFAULT, a=1, b=2)
    assert set(ttl_dict.keys()) == {"a", "b"}
    time.sleep(SLEEP_DEFAULT)
    assert len(ttl_dict) == 0
    assert set(ttl_dict.keys()) == set()


def test_values():
    ttl_dict = TTLDict(TTL_DEFAULT)
    orig_dict = {"a": 1, "b": 2}
    ttl_dict.update(orig_dict)
    values = set(ttl_dict.values())
    assert len(values) == 2
    assert values == {1, 2}


def test_items():
    ttl_dict = TTLDict(TTL_DEFAULT)
    orig_dict = {"a": 1, "b": 2}
    ttl_dict.update(orig_dict)
    items = list(ttl_dict.items())
    assert len(items) == 2
    assert items == [("a", 1), ("b", 2)]


def test_ttl_dict_from_multiple_threads():
    ttl = 0.2
    ttl_dict = TTLDict(ttl)
    threads = []
    thread_count = 1000

    def lazy_set(n: int):
        ttl_dict[n] = n
        time.sleep(0.05)

    for i in range(thread_count):
        thread = Thread(target=lazy_set, args=(i,))
        threads.append(thread)
    for thread in threads:
        thread.start()
    assert len(ttl_dict) > 0
    for thread in threads:
        thread.join()

    assert len(ttl_dict) == thread_count
    for i in range(thread_count):
        assert ttl_dict[i] == i

    time.sleep(ttl)
    assert len(ttl_dict) == 0



import logging
from typing import Optional

from the_spymaster_util.http.defs import ErrorTypes, JsonType

log = logging.getLogger(__name__)


class ApiError(Exception):
    def __init__(
        self,
        message: str = "Internal server error",
        status_code: int = 500,
        details: Optional[JsonType] = None,
    ):
        self.message = message
        self.status_code = status_code
        self.details = details
        super().__init__(message)


class BadRequestError(ApiError):
    def __init__(self, message: str, status_code: int = 400, details: Optional[JsonType] = None):
        super().__init__(message=message, status_code=status_code, details=details)


class UnauthorizedError(BadRequestError):
    def __init__(self, message: str):
        super().__init__(message=message, status_code=401)


class ForbiddenError(BadRequestError):
    def __init__(self, message: str):
        super().__init__(message=message, status_code=403)


class NotFoundError(BadRequestError):
    def __init__(self, message: str):
        super().__init__(message=message, status_code=404)


DEFAULT_ERRORS: ErrorTypes = frozenset({ApiError, BadRequestError, UnauthorizedError, ForbiddenError, NotFoundError})



import threading
from queue import Empty
from random import random
from time import sleep

import pytest

from the_spymaster_util.async_task_manager import AsyncTaskManager
from the_spymaster_util.logger import get_logger

log = get_logger(__name__)


def dummy_task(x: int, duration: float = None):
    log.update_thread_context(x=f"{x}")
    if not duration:
        duration = random() / 20
    sleep(duration)
    log.debug(f"Im done with {x}.")
    return x


def test_manager_is_results_iterable():
    manager = AsyncTaskManager(workers_amount=5)
    inputs = set(range(20))
    for i in inputs:
        manager.add_task(dummy_task, args=(i,))

    manager.join()
    results = set(manager)

    assert results == inputs
    assert manager.is_work_done
    assert manager.is_empty


def test_manager_is_a_context_manager():
    with AsyncTaskManager(workers_amount=5) as manager:
        inputs = set(range(20))
        for i in inputs:
            manager.add_task(dummy_task, args=(i,))
        results = set(manager)

        assert results == inputs
        assert manager.is_work_done
        assert manager.is_empty
        assert not manager.is_closed
    assert manager.is_closed


def test_manager_raises_empty_exception_if_no_results_after_timeout():
    manager = AsyncTaskManager(workers_amount=1)
    with pytest.raises(Empty):
        _ = manager.get_result(timeout=0.05)
    manager.join()


def test_no_tasks_are_processed_after_manager_joined():
    manager = AsyncTaskManager(workers_amount=5)
    inputs = set(range(50))
    for i in inputs:
        manager.add_task(dummy_task, args=(i,))
    manager.join()
    assert manager.is_closed

    manager.add_task(dummy_task, args=(999,))
    assert manager.is_work_done
    assert not manager.is_empty
    results = set(manager)
    assert manager.is_empty
    assert results == inputs


def test_no_threads_are_left_open_after_manager_completes():
    active_threads = threading.active_count()
    manager = AsyncTaskManager(workers_amount=5)
    inputs = set(range(50))
    for i in inputs:
        manager.add_task(dummy_task, args=(i,))
    results = set(manager)
    assert results == inputs
    assert manager.is_closed
    sleep(0.1)
    assert threading.active_count() == active_threads

    manager.add_task(dummy_task, args=(999,))
    assert manager.is_work_done
    assert manager.is_empty



import re
from typing import Type, TypeVar, Union

CAMEL_TO_CONST_1 = re.compile(r"(.)([A-Z][a-z]+)")
CAMEL_TO_CONST_2 = re.compile(r"([a-z0-9])([A-Z])")
T = TypeVar("T")


def camel_to_const(string: str) -> str:
    string = CAMEL_TO_CONST_1.sub(r"\1_\2", string)
    return CAMEL_TO_CONST_2.sub(r"\1_\2", string).upper()


def get_class_name(instance_or_class: Union[T, Type[T]]):
    clazz = instance_or_class if isinstance(instance_or_class, type) else instance_or_class.__class__
    return clazz.__name__



import json
import logging
import sys
import threading
from datetime import datetime, timezone
from logging import Filter, Formatter, Logger, LogRecord
from typing import Any, Dict, Optional

import ulid

ContextDict = Dict[str, Any]
CONTEXT_KEY = "_thread_logging_context"

_thread_storage = threading.local()
_process_context: ContextDict = {}


class ContextLogger(Logger):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._thread_storage = _thread_storage

    @property
    def context_id(self) -> Optional[str]:
        return self.context.get("context_id")

    # Process context

    @property
    def context(self) -> ContextDict:
        if not _process_context:
            self.reset_context()
        return {**_process_context, **self.thread_context}

    @classmethod
    def update_context(cls, **kwargs) -> ContextDict:
        _process_context.update(kwargs)
        return _process_context

    @classmethod
    def set_context(cls, context: Optional[ContextDict] = None, **kwargs) -> ContextDict:
        context = context or {}
        if "context_id" not in context:
            context = {"context_id": ulid.new().str, **context}  # Always keep context_id first.
        _process_context.clear()
        return cls.update_context(**context, **kwargs)

    @classmethod
    def reset_context(cls) -> ContextDict:
        return cls.set_context({})

    # Thread context

    @property
    def thread_context(self) -> ContextDict:
        current_context = getattr(self._thread_storage, CONTEXT_KEY, None)
        if current_context is None:
            return self.reset_thread_context()
        return current_context

    def update_thread_context(self, **kwargs) -> ContextDict:
        new_context = self.thread_context
        new_context.update(kwargs)
        return self.set_thread_context(new_context)

    def set_thread_context(self, context: ContextDict) -> ContextDict:
        setattr(self._thread_storage, CONTEXT_KEY, context)
        return context

    def reset_thread_context(self) -> ContextDict:
        return self.set_thread_context({})

    # Logging

    def _log(self, *args, **kwargs) -> None:
        extra = kwargs.get("extra")
        kwargs["extra"] = {"extra": extra, "context": self.context}
        super()._log(*args, **kwargs)  # noqa


class ContextFormatter(Formatter):
    def __init__(self, *args, log_extra: bool = True, log_context: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.log_extra = log_extra
        self.log_context = log_context

    def format(self, record: LogRecord) -> str:
        message = original_message = record.msg
        extra = getattr(record, "extra", None)
        context = getattr(record, "context", None)
        message = str(message)
        if extra and self.log_extra:
            message += f" | {extra}"
        if context and self.log_context:
            message += f" | {context}"
        record.msg = message
        result = super().format(record)
        record.msg = original_message
        return result


class JsonFormatter(Formatter):
    def __init__(self, *args, indented: bool = False, tz: Optional[timezone] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.indent = 2 if indented else None
        self.tz = tz or datetime.now().astimezone().tzinfo

    def format(self, record: LogRecord) -> str:
        record_data = {
            # Base
            "date_time": self._format_date_time(record.created),
            "level": record.levelname,
            "message": super().format(record),
            "extra": getattr(record, "extra", None),
            "context": getattr(record, "context", None),
            # Debug
            "func_name": record.funcName,
            "file_path": record.pathname,
            "line_number": record.lineno,
            # More
            "level_code": record.levelno,
            "timestamp": record.created,
            "logger": record.name,
            "thread_name": record.threadName,
        }
        try:
            return json.dumps(record_data, indent=self.indent, ensure_ascii=False)
        except Exception as e:  # noqa
            log.debug(f"Record serialization failed: {e}")
            return str(record_data)

    def _format_date_time(self, timestamp: float) -> str:
        return datetime.fromtimestamp(timestamp, self.tz).isoformat(sep=" ", timespec="milliseconds")


class LevelRangeFilter(Filter):
    def __init__(self, low=0, high=100):
        Filter.__init__(self)
        self.low = low
        self.high = high

    def filter(self, record):
        if self.low <= record.levelno < self.high:
            return True
        return False


def get_logger(name: str) -> ContextLogger:
    return logging.getLogger(name)  # type: ignore


def wrap(o: object) -> str:
    return f"[{o}]"


def get_dict_config(
    *,
    std_formatter: str = "json",
    root_log_level: str = "DEBUG",
    indent_json: bool = False,
    extra_handlers: dict = None,
    extra_loggers: dict = None,
) -> dict:
    dict_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {
                "()": "the_spymaster_util.logger.ContextFormatter",
                "format": "[%(asctime)s.%(msecs)03d] [%(levelname)-.4s]: %(message)s [%(name)s]",
                "datefmt": "%H:%M:%S",
                "log_context": False,
            },
            "debug": {
                "class": "the_spymaster_util.logger.ContextFormatter",
                "format": "[%(asctime)s.%(msecs)03d] [%(levelname)-.4s]: %(message)s @@@ "
                "[%(name)s:%(lineno)s] [%(threadName)s]",
                "datefmt": "%Y-%m-%d %H:%M:%S",
            },
            "json": {
                "()": "the_spymaster_util.logger.JsonFormatter",
                "indented": indent_json,
            },
        },
        "filters": {
            "std_filter": {"()": "the_spymaster_util.logger.LevelRangeFilter", "high": logging.WARNING},
            "err_filter": {"()": "the_spymaster_util.logger.LevelRangeFilter", "low": logging.WARNING},
        },
        "handlers": {
            "console_out": {
                "class": "logging.StreamHandler",
                "filters": ["std_filter"],
                "formatter": std_formatter,
                "stream": sys.stdout,
            },
            "console_err": {
                "class": "logging.StreamHandler",
                "filters": ["err_filter"],
                "formatter": std_formatter,
                "stream": sys.stderr,
            },
        },
        "root": {"handlers": ["console_out", "console_err"], "level": root_log_level},
        "loggers": {},
    }
    if extra_handlers:
        dict_config["handlers"].update(extra_handlers)  # type: ignore
    if extra_loggers:
        dict_config["loggers"].update(extra_loggers)  # type: ignore
    return dict_config


logging.setLoggerClass(ContextLogger)
log = get_logger(__name__)



from logging.config import dictConfig
from random import random
from threading import Thread
from time import sleep

from the_spymaster_util.logger import get_dict_config, get_logger


def configure_logging():
    dict_config = get_dict_config()
    dictConfig(dict_config)


configure_logging()
log = get_logger(__name__)


def worker(name: str):
    log.update_context(thread=name)
    log.info(f"{name} start")
    sleep(random() / 10)
    log.info(f"{name} end  ")


def main():
    # TODO: Make this a test.
    log.update_context(shared=True)
    log.info("main")
    for i in range(5):
        t = Thread(target=worker, args=(f"t{i + 1}",), name=f"t{i + 1}")
        t.start()


if __name__ == "__main__":
    main()



import base64
import json
import logging
import os
from typing import Any, List, Optional

import boto3
from dynaconf import Dynaconf

log = logging.getLogger(__name__)


class LazyConfig:
    def __init__(self, settings: Dynaconf = None):
        self._settings = settings

    def __getattr__(self, item) -> Any:
        return self.get(item)

    def __getitem__(self, item):
        return self.get(item)

    @property
    def settings(self) -> Dynaconf:
        if not self._settings:
            self.load()
        return self._settings

    @property
    def env_name(self) -> Optional[str]:
        return self.get("ENV_FOR_DYNACONF")

    def get(self, key: str, default=None) -> Any:
        return self.settings.get(key) or os.getenv(key) or default

    def set(self, key: str, value: Any):
        if not self._settings:
            raise Exception("Config not loaded")
        self._settings.set(key, value)

    def update(self, **kwargs):
        if not self._settings:
            raise Exception("Config not loaded")
        self._settings.update(**kwargs)

    def load(self, extra_files: List[str] = None):
        log.info("Loading configurations...")
        if not extra_files:
            extra_files = []
        settings_files = ["settings.toml", "local.toml", "secrets.toml"] + extra_files
        self._settings = Dynaconf(environments=True, settings_files=settings_files)

    def load_from_secrets_manager(self, secret_id: Optional[str]) -> dict:
        if not secret_id:
            log.info("No secret_id provided")
            return {}
        client = boto3.client(service_name="secretsmanager")
        response = client.get_secret_value(SecretId=secret_id)
        if "SecretString" in response:
            secrets_string = response["SecretString"]
        else:
            secrets_string = base64.b64decode(response["SecretBinary"])
        secrets_dict = json.loads(secrets_string) or {}
        self.update(**secrets_dict)  # type: ignore
        log.info("Secrets loaded successfully")
        return secrets_dict

    def load_ssm_parameters(self, names: List[str]):
        ssm_client = boto3.client("ssm")
        response = ssm_client.get_parameters(Names=names, WithDecryption=True)
        for param in response["Parameters"]:
            name, value = param["Name"], param["Value"]
            self.set(name, value)
        for param in response["InvalidParameters"]:
            log.warning(f"Invalid parameter: {param}")



from http.client import HTTPMessage
from typing import Optional
from unittest.mock import ANY, Mock, call, patch

from urllib3 import Retry

from the_spymaster_util.http.base_client import BaseHttpClient


class ExampleClient(BaseHttpClient):
    def __init__(self, retry_strategy: Optional[Retry] = None):
        super().__init__(base_url="https://jsonplaceholder.typicode.com", retry_strategy=retry_strategy)

    def users(self, **kwargs):
        return self._get("users", {}, **kwargs)

    def dummy(self):
        return self._get("dummy", data={}, parse=False)


def test_basic_call_succeed():
    client = ExampleClient()
    response = client.users()
    assert isinstance(response, list)
    assert len(response) == 10


def test_basic_call_without_log_data():
    client = ExampleClient()
    response = client.users(log_http_data=False)
    assert isinstance(response, list)
    assert len(response) == 10


@patch("urllib3.connectionpool.HTTPConnectionPool._get_conn")
def test_retry_request(mock_get_conn):
    mock_get_conn.return_value.getresponse.side_effect = [
        Mock(status=429, msg=HTTPMessage()),  # Retry #0 (initial)
        Mock(status=429, msg=HTTPMessage()),  # Retry #1
        Mock(status=500, msg=HTTPMessage()),  # Retry #2
        Mock(status=200, msg=HTTPMessage()),  # Retry #3 (last, success)
        Mock(status=200, msg=HTTPMessage()),  # We won't get to this one.
    ]

    retry_strategy = Retry(
        total=3,
        backoff_factor=0.01,
        status_forcelist=[429, 500, 502, 503, 504],  # The HTTP response codes to retry on
        allowed_methods=["HEAD", "GET", "PUT", "DELETE", "OPTIONS"],  # The HTTP methods to retry on
    )
    client = ExampleClient(retry_strategy=retry_strategy)
    client.dummy()

    actual_calls = mock_get_conn.return_value.request.mock_calls
    assert actual_calls == [
        call("GET", "/dummy", body=None, headers=ANY),
        call("GET", "/dummy", body=None, headers=ANY),
        call("GET", "/dummy", body=None, headers=ANY),
        call("GET", "/dummy", body=None, headers=ANY),
    ]
